//BufferManager.java

import java.nio.*;
import java.util.*;
/**
    This class is used to manage reading and writing pages on a
    MockDisk object. All read requests from a MockDiskshould go
    through yhe drive's BufferManager object: driveBuffer.
    To get a reference to the page numbered i, one calls

    Page p = driveBuffer.readPage(i);

    This pins the page in the BufferManager's buffer pool. So
    when one is done with the page one has to call

    driveBuffer.releasePage(i);

    If one has a reference to a page via the buffer and one wants
    to change the page then it is the user of the pages responsibility
    to set the dirty bit of the page so that the BufferManager knows
    to write the page back to disk if it gets replaced from the buffer.
    To set the dirty bit one does:

    driveBuffer.setDirty(i);

    BEFORE releasing the page.

    @author Chris Pollett
    @version 2003/10/6
*/

public class BufferManager
{
   /**
         Empty constructor. Set-up of class deferred to create method below
   */
   public BufferManager()
   {
   }

   /**
      Sets up the BufferManager creates the bufferPool,
      and arryas to keep track pin count, dirtiness and which
      pages are in the pool. Sets the passed drive full of free
      pages to be the pages managed by this BufferManager.

      @param size - how many pages the buffer pool will be
      @param drive - pages used to represent the disk drive
    */
   public void create(int size, Page[] drive)
   {
      bufferSize = size;
      bufferPool = new Page[size];
      pinCount = new int[size];
      pageNumbers = new int[size];
      dirty = new boolean[size];
      this.drive=drive;

      storedPages= new Hashtable();

      driveSize = drive.length;
      clock = 0;

      for(int i = 0; i < size; i++)
      {
         bufferPool[i] = new Page();
      }
   }

   /**
      Tries to find requested page in buffer pool and if so
      updates the pin count of the page and returns a
      reference to it. If not it uses getPage to swap the page
      into the buffer pool.

      @param pageNumber - which page of drive to return

      @throws IndexOutOfBoundsException - thrown if pageNumber not in drive

      @return - requested page
   */
   public Page readPage(int pageNumber) throws IndexOutOfBoundsException
   {
      isValidNumber(pageNumber, 0, driveSize);
      Object indexObj=storedPages.get(new Integer(pageNumber));
      int index;
      if(indexObj == null)
      {
         index = getPage(pageNumber);
         pinCount[index] = 1;
      }
      else
      {
         index = ((Integer)indexObj).intValue();
         pinCount[index]++;
      }

      return bufferPool[index];
   }
   /**
      Sets the dirty bit of the indicated page. The dirty bit being
      set indicates that if the page is removed from the buffer
      it must be first written back to disk.

      @param pageNumber - page whose dirty bit needs to be set

      @throws MissingResourceException - thrown if page not currently in buffer
    */
   public void setDirty(int pageNumber) throws MissingResourceException
   {
      isValidNumber(pageNumber, 0, driveSize);
      Object indexObj=storedPages.get(new Integer(pageNumber));
      if(indexObj == null)
         throw new MissingResourceException("Page not in buffer pool.",
            "BufferManager", ""+pageNumber);

      dirty[((Integer)indexObj).intValue()] = true;

   }


   /**
      Releases one pin on the given page in the buffer pool.

      @param pageNumber - page to release a pin of.
      @throws IllegalStateException - if pin count on a page goes negative.
    */
   public void releasePage(int pageNumber)
      throws IllegalStateException

   {
      isValidNumber(pageNumber, 0, driveSize);

      Object indexObj = storedPages.get(new Integer(pageNumber));
      if(indexObj == null)
         throw new MissingResourceException("Page not in buffer pool.",
            "BufferManager", ""+pageNumber);

      int index =((Integer)indexObj).intValue();
      if(pinCount[index] <= 0)
         throw new IllegalStateException("Page "+pageNumber
            + "released too many times");

      pinCount[index]--;


   }

   /**
      Gets a page that is not currently in the buffer pool.
      If the page that is replaced has its dirty bit
      set then it is written back to the disk. This
      method uses the clock replacement policy to determine
      which should be the next buffer page to be used
      to read disk page into. For the Homework you should
      override this method to use the LRU policy.

      @param pageNumber - page to retrieve.

      @throws BufferOverflowException - if page is not in buffer and
          all buffer pages are currently pinned.
    */
   int getPage(int pageNumber)
      throws BufferOverflowException
   {
      isValidNumber(pageNumber, 0, driveSize);
      int index = clock + 1;
      index = (index < bufferSize) ? index : 0;
      while( index != clock && pinCount[index] > 0)
      {
         index++;
         index = (index < bufferSize) ? index : 0;
      }

      if( pinCount[index] > 0)
         throw new BufferOverflowException();

      if(dirty[index])
      {
         drive[pageNumbers[index]].copy(bufferPool[index]);
         dirty[index] = false;
      }

      storedPages.remove(new Integer(pageNumbers[index]));
      bufferPool[index].copy(drive[pageNumber]);
      pageNumbers[index] = pageNumber;
      storedPages.put(new Integer(pageNumber), new Integer(index));
      pinCount[index] = 1;

      clock++;
      clock = (clock < bufferSize) ? clock : 0;

      return index;
   }

   /**
      Generates a String to represent the data stored about a given page
      in the buffer pool. i.e., which disk page, its pin count,
      its dirtyness and its data.

      @return the string with this information
   */
   public String toStringBufferPoolPage(int i)
   {
      isValidNumber(i, 0, bufferSize-1);
      String out = "";
      out += "Buffer Page: " + i + " pageNumber: "+ pageNumbers[i];
      out += " pinCount: "+ pinCount[i] + " dirty: " + dirty[i] +"\n";

      out += "\t"+bufferPool[i].toStringDataPage();
      return out;

   }

   /**
      Checks if number is <=high and >= low. If not throws an
      IndexOutofBoundsException. Method says which class it belongs
      to (maybe could have avoided code duplication)

      @param number - number to check
      @param low - low value of range number should be in
      @param high - high value of range number should be in
      @throws IndexOutOfBoundsException - returned if number not in range
    */
   static void isValidNumber(int number, int low, int high)
      throws IndexOutOfBoundsException
   {
      if(number > high || number < low)
         throw new IndexOutOfBoundsException("(Buffer Index Exception)" +
            "Number was:" + number +".\n Maximum allowable is: " + high +
            ".\n Minimum allowable is: " + low + ".\n");
   }

   int driveSize; // size in number of pages in the drive

   int bufferSize; // size in number of pages in the buffer pool

   int clock; /*keeps track of where the clock is at in terms of cycling
                through pages in the buffer pool for clock replacement
                buffer policy */

   Page[] bufferPool; // pages used for the buffer

   Page[] drive; // the actual pages in the disk drive

   int[] pinCount; // used to keep track of the number of times a
                   // page has been pinned

   int[] pageNumbers; // used to keep track of which page a given
                      // buffer stores

   boolean[] dirty; // used to keep track of which buffer pages are dirty

   Hashtable storedPages; // used to quickly look-up via drive page number
                          // where a give page is stored in the buffer pool
}

